@@ -5,7 +5,7 @@
 %%% Created :  8 Aug 2009 by Evgeniy Khramtsov <ekhramtsov@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2021 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2023 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -54,6 +54,7 @@
 
 -define(MAX_BUF_SIZE, 64*1024). %% 64kb
 -define(TIMEOUT, 60000). %% 1 minute
+-define(TCP_ACTIVE, 500).
 -define(NONCE_LIFETIME, 60*1000*1000). %% 1 minute (in usec)
 -define(SERVER_NAME, <<"P1 STUN library">>).
 
@@ -62,7 +63,6 @@
 -record(state,
 	{sock                        :: inet:socket() | fast_tls:tls_socket() | undefined,
 	 sock_mod = gen_tcp          :: gen_udp | gen_tcp | fast_tls,
-	 certfile                    :: iodata() | undefined,
 	 peer = {{0,0,0,0}, 0}       :: addr(),
 	 tref                        :: reference() | undefined,
 	 use_turn = false            :: boolean(),
@@ -73,13 +73,16 @@
 	 max_allocs = 10             :: non_neg_integer() | infinity,
 	 shaper = none               :: stun_shaper:shaper(),
 	 max_permissions = 10        :: non_neg_integer() | infinity,
-	 blacklist = []              :: turn:blacklist(),
+	 blacklist_clients = []      :: turn:accesslist(),
+	 whitelist_clients = []      :: turn:accesslist(),
+	 blacklist_peers = []        :: turn:accesslist(),
+	 whitelist_peers = []        :: turn:accesslist(),
 	 auth = user                 :: anonymous | user,
 	 nonces = treap:empty()      :: treap:treap(),
 	 realm = <<"">>              :: binary(),
 	 auth_fun                    :: function() | undefined,
 	 hook_fun                    :: function() | undefined,
-	 server_name = ?SERVER_NAME  :: binary(),
+	 server_name = ?SERVER_NAME  :: binary() | undefined,
 	 buf = <<>>                  :: binary(),
 	 session_id                  :: binary() | undefined}).
 
@@ -107,11 +110,11 @@ udp_init(Sock, Opts) ->
 udp_recv(Sock, Addr, Port, Data, State) ->
     NewState = prepare_state(State, Sock, {Addr, Port}, gen_udp),
     case stun_codec:decode(Data, datagram) of
- 	{ok, Msg} ->
- 	    ?LOG_DEBUG(#{verbatim => {"Received:~n~s", [stun_codec:pp(Msg)]}}),
- 	    process(NewState, Msg);
- 	Err ->
-	    ?LOG_DEBUG("Cannot parse packet: ~p", [Err]),
+	{ok, Msg} ->
+	    ?LOG_DEBUG(#{verbatim => {"Received:~n~s", [stun_codec:pp(Msg)]}}),
+	    process(NewState, Msg);
+	{error, Reason} ->
+	    ?LOG_DEBUG("Cannot parse packet: ~s", [Reason]),
 	    NewState
     end.
 
@@ -120,26 +123,23 @@ udp_recv(Sock, Addr, Port, Data, State) ->
 %%====================================================================
 init([Sock, Opts]) ->
     process_flag(trap_exit, true),
-    case inet:peername(Sock) of
+    case get_peername(Sock, Opts) of
 	{ok, Addr} ->
 	    case get_sockmod(Opts, Sock) of
 		{ok, SockMod} ->
 		    State = prepare_state(Opts, Sock, Addr, SockMod),
-		    CertFile = get_certfile(Opts),
-		    case maybe_starttls(Sock, SockMod, CertFile) of
+		    case maybe_starttls(Sock, SockMod, Opts) of
 			{ok, NewSock} ->
-			    inet:setopts(Sock, [{active, once}]),
 			    TRef = erlang:start_timer(?TIMEOUT, self(), stop),
-			    {ok, session_established,
-			     State#state{tref = TRef, sock = NewSock}};
+			    NewState = State#state{sock = NewSock, tref = TRef},
+			    activate_socket(NewState),
+			    {ok, session_established, NewState};
 			{error, Reason} ->
 			    {stop, Reason}
-		    end;
-		{error, Reason} ->
-		    {stop, Reason}
+		    end
 	    end;
-	Err ->
-	    {stop, Err}
+	{error, Reason} ->
+	    {stop, Reason}
     end.
 
 session_established(Event, State) ->
@@ -160,13 +160,16 @@ handle_info({tcp, _Sock, TLSData}, StateName,
     case fast_tls:recv_data(NewState#state.sock, TLSData) of
 	{ok, Data} ->
 	    process_data(StateName, NewState, Data);
-	Err ->
-	    ?LOG_INFO("Connection failure: ~p", [Err]),
+	{error, Reason} ->
+	    ?LOG_INFO("Connection failure: ~s", [Reason]),
 	    {stop, normal, NewState}
     end;
 handle_info({tcp, _Sock, Data}, StateName, State) ->
     NewState = update_shaper(State, Data),
     process_data(StateName, NewState, Data);
+handle_info({tcp_passive, _Sock}, StateName, State) ->
+    activate_socket(State),
+    {next_state, StateName, State};
 handle_info({tcp_closed, _Sock}, _StateName, State) ->
     ?LOG_INFO("Connection reset by peer"),
     {stop, normal, State};
@@ -222,6 +225,7 @@ process(#state{auth = anonymous} = State,
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(401)},
+    run_hook(protocol_error, State, R),
     send(State, R);
 process(#state{auth = user} = State,
 	#stun{class = request,
@@ -242,16 +246,19 @@ process(#state{auth = user} = State,
 	    case (State#state.auth_fun)(User, Realm) of
 		<<"">> ->
 		    ?LOG_NOTICE("Failed long-term STUN/TURN authentication"),
+		    run_hook(protocol_error, State, R),
 		    send(NewState, R);
-		Pass ->
+		Pass0 ->
+		    {Pass, IsExpired} = check_expired_tag(Pass0),
 		    case check_integrity(User, Realm, Msg, Pass) of
 			{true, Key} ->
 			    ?LOG_INFO("Accepting long-term STUN/TURN "
 				      "authentication"),
-			    process(NewState, Msg, Key);
+			    process(NewState, Msg, Key, IsExpired);
 			false ->
 			    ?LOG_NOTICE("Failed long-term STUN/TURN "
 					"authentication"),
+			    run_hook(protocol_error, State, R),
 			    send(NewState, R)
 		    end
 	    end;
@@ -262,6 +269,7 @@ process(#state{auth = user} = State,
 			  'ERROR-CODE' = stun_codec:error(438),
 			  'REALM' = State#state.realm,
 			  'NONCE' = NewNonce},
+	    run_hook(protocol_error, State, R),
 	    send(State#state{nonces = NewNonces}, R)
     end;
 process(State, #stun{class = request,
@@ -272,12 +280,14 @@ process(State, #stun{class = request,
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(401)},
+    run_hook(protocol_error, State, R),
     send(State, R);
 process(State, #stun{class = request} = Msg) ->
     ?LOG_NOTICE("Rejecting malformed request"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(400)},
+    run_hook(protocol_error, State, R),
     send(State, R);
 process(State, #stun{class = indication,
 		     method = ?STUN_METHOD_SEND} = Msg) ->
@@ -287,17 +297,22 @@ process(State, Msg) when is_record(Msg, turn) ->
 process(State, _Msg) ->
     State.
 
+process(State, Msg, Secret) ->
+    process(State, Msg, Secret, false).
+
 process(State, #stun{class = request, unsupported = [_|_] = Unsupported} = Msg,
-	Secret) ->
+	Secret, _IsExpired) ->
     ?LOG_DEBUG("Rejecting request with unknown attribute(s): ~p",
 	       [Unsupported]),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'UNKNOWN-ATTRIBUTES' = Unsupported,
 		  'ERROR-CODE' = stun_codec:error(420)},
+    run_hook(protocol_error, State, R),
     send(State, R, Secret);
 process(State, #stun{class = request,
-		     method = ?STUN_METHOD_BINDING} = Msg, Secret) ->
+		     method = ?STUN_METHOD_BINDING} = Msg, Secret,
+	_IsExpired) ->
     Resp = prepare_response(State, Msg),
     AddrPort = unmap_v4_addr(State#state.peer),
     R = case stun_codec:version(Msg) of
@@ -311,14 +326,15 @@ process(State, #stun{class = request,
     run_hook(stun_query, State, Msg),
     send(State, R, Secret);
 process(#state{use_turn = false} = State,
-	#stun{class = request} = Msg, Secret) ->
+	#stun{class = request} = Msg, Secret, _IsExpired) ->
     ?LOG_NOTICE("Rejecting TURN request: TURN is disabled"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(405)},
+    run_hook(protocol_error, State, R),
     send(State, R, Secret);
 process(State, #stun{class = request,
 		     method = ?STUN_METHOD_ALLOCATE} = Msg,
-	Secret) ->
+	Secret, IsExpired) ->
     Resp = prepare_response(State, Msg),
     AddrPort = State#state.peer,
     SockMod = State#state.sock_mod,
@@ -326,16 +342,24 @@ process(State, #stun{class = request,
 	{ok, Pid} ->
 	    turn:route(Pid, Msg),
 	    State;
+	_ when IsExpired ->
+	    ?LOG_NOTICE("Rejecting request: credentials expired"),
+	    R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(401)},
+	    run_hook(protocol_error, State, R),
+	    send(State, R);
 	_ ->
 	    Opts = [{sock, State#state.sock},
 		    {sock_mod, SockMod},
 		    {username, Msg#stun.'USERNAME'},
 		    {realm, State#state.realm},
 		    {key, Secret},
-		    {server, State#state.server_name},
+		    {server_name, State#state.server_name},
 		    {max_allocs, State#state.max_allocs},
 		    {max_permissions, State#state.max_permissions},
-		    {blacklist, State#state.blacklist},
+		    {blacklist_clients, State#state.blacklist_clients},
+		    {whitelist_clients, State#state.whitelist_clients},
+		    {blacklist_peers, State#state.blacklist_peers},
+		    {whitelist_peers, State#state.whitelist_peers},
 		    {addr, AddrPort},
 		    {relay_ipv4_ip, State#state.relay_ipv4_ip},
 		    {relay_ipv6_ip, State#state.relay_ipv6_ip},
@@ -357,32 +381,39 @@ process(State, #stun{class = request,
 		    ?LOG_NOTICE("Rejecting request: Allocation quota reached"),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(486)},
+		    run_hook(protocol_error, State, R),
 		    send(State, R, Secret);
 		{error, stale} ->
 		    ?LOG_NOTICE("Rejecting request: Stale nonce"),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(438)},
+		    run_hook(protocol_error, State, R),
 		    send(State, R);
-		Err ->
-		    ?LOG_ERROR("Cannot start TURN session: ~p", [Err]),
+		{error, Reason} ->
+		    ?LOG_ERROR("Cannot start TURN session: ~s", [Reason]),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(500)},
+		    run_hook(protocol_error, State, R),
 		    send(State, R, Secret)
 	    end
     end;
 process(State, #stun{class = request,
-		     method = ?STUN_METHOD_REFRESH} = Msg, Secret) ->
+		     method = ?STUN_METHOD_REFRESH} = Msg, Secret,
+	_IsExpired) ->
     route_on_turn(State, Msg, Secret);
 process(State, #stun{class = request,
-		     method = ?STUN_METHOD_CREATE_PERMISSION} = Msg, Secret) ->
+		     method = ?STUN_METHOD_CREATE_PERMISSION} = Msg, Secret,
+	_IsExpired) ->
     route_on_turn(State, Msg, Secret);
 process(State, #stun{class = request,
-		    method = ?STUN_METHOD_CHANNEL_BIND} = Msg, Secret) ->
+		    method = ?STUN_METHOD_CHANNEL_BIND} = Msg, Secret,
+	_IsExpired) ->
     route_on_turn(State, Msg, Secret);
-process(State, #stun{class = request} = Msg, Secret) ->
+process(State, #stun{class = request} = Msg, Secret, _IsExpired) ->
     ?LOG_NOTICE("Rejecting request: Method not allowed"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(405)},
+    run_hook(protocol_error, State, R),
     send(State, R, Secret).
 
 process_data(NextStateName, #state{buf = Buf} = State, Data) ->
@@ -403,6 +434,8 @@ process_data(NextStateName, #state{buf = Buf} = State, Data) ->
 	    {stop, normal, State}
     end.
 
+update_shaper(#state{shaper = none} = State, _Data) ->
+    State;
 update_shaper(#state{shaper = Shaper} = State, Data) ->
     {NewShaper, Pause} = stun_shaper:update(Shaper, size(Data)),
     if Pause > 0 ->
@@ -448,6 +481,7 @@ route_on_turn(State, Msg, Pass) ->
 		    Resp = prepare_response(State, Msg),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(437)},
+		    run_hook(protocol_error, State, R),
 		    send(State, R, Pass);
 		_ ->
 		    State
@@ -520,21 +554,78 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 		      ?LOG_ERROR("Wrong 'turn_max_permissions' value: ~p",
 				 [Wrong]),
 		      State;
-		 ({turn_blacklist, B}, State) ->
+		 ({turn_blacklist_clients, B},
+		  #state{blacklist_clients = B0} = State) ->
 		      case lists:all(fun is_valid_subnet/1, B) of
 			  true ->
-			      State#state{blacklist = B};
+			      State#state{blacklist_clients = B0 ++ B};
 			  false ->
-			      ?LOG_ERROR("Wrong 'turn_blacklist' value: ~p",
-					 [B]),
+			      ?LOG_ERROR("Wrong 'turn_blacklist_clients' "
+					 "value: ~p", [B]),
 			      State
 		      end;
+		 ({turn_whitelist_clients, W},
+		  #state{whitelist_clients = W0} = State) ->
+		      case lists:all(fun is_valid_subnet/1, W) of
+			  true ->
+			      State#state{whitelist_clients = W0 ++ W};
+			  false ->
+			      ?LOG_ERROR("Wrong 'turn_whitelist_clients' "
+					 "value: ~p", [W]),
+			      State
+		      end;
+		 ({turn_blacklist_peers, B},
+		  #state{blacklist_peers = B0} = State) ->
+		      case lists:all(fun is_valid_subnet/1, B) of
+			  true ->
+			      State#state{blacklist_peers = B0 ++ B};
+			  false ->
+			      ?LOG_ERROR("Wrong 'turn_blacklist_peers' "
+					 "value: ~p", [B]),
+			      State
+		      end;
+		 ({turn_whitelist_peers, W},
+		  #state{whitelist_peers = W0} = State) ->
+		      case lists:all(fun is_valid_subnet/1, W) of
+			  true ->
+			      State#state{whitelist_peers = W0 ++ W};
+			  false ->
+			      ?LOG_ERROR("Wrong 'turn_whitelist_peers' "
+					 "value: ~p", [W]),
+			      State
+		      end;
+		 ({turn_blacklist, B}, #state{blacklist_clients = C,
+					      blacklist_peers = P} = State0) ->
+		      case lists:all(fun is_valid_subnet/1, B) of
+			  true ->
+			      State1 = State0#state{blacklist_clients = C ++ B},
+			      State2 = State1#state{blacklist_peers = P ++ B},
+			      State2;
+			  false ->
+			      ?LOG_ERROR("Wrong 'turn_blacklist' "
+					 "value: ~p", [B]),
+			      State0
+		      end;
+		 ({turn_whitelist, W}, #state{whitelist_clients = C,
+					      whitelist_peers = P} = State0) ->
+		      case lists:all(fun is_valid_subnet/1, W) of
+			  true ->
+			      State1 = State0#state{whitelist_clients = C ++ W},
+			      State2 = State1#state{whitelist_peers = P ++ W},
+			      State2;
+			  false ->
+			      ?LOG_ERROR("Wrong 'turn_whitelist' "
+					 "value: ~p", [W]),
+			      State0
+		      end;
 		 ({shaper, S}, State)
 		    when S == none orelse (is_integer(S) andalso (S > 0)) ->
 		      State#state{shaper = stun_shaper:new(S)};
 		 ({shaper, Wrong}, State) ->
 		      ?LOG_ERROR("Wrong 'shaper' value: ~p", [Wrong]),
 		      State;
+		 ({server_name, none}, State) ->
+		      State#state{server_name = undefined};
 		 ({server_name, S}, State) ->
 		      try
 			  State#state{server_name = iolist_to_binary(S)}
@@ -572,8 +663,14 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 		 ({ip, _}, State) -> State;
 		 ({backlog, _}, State) -> State;
 		 ({certfile, _}, State) -> State;
+		 ({dhfile, _}, State) -> State;
+		 ({ciphers, _}, State) -> State;
+		 ({protocol_options, _}, State) -> State;
 		 ({tls, _}, State) -> State;
 		 (tls, State) -> State;
+		 ({proxy_protocol, _}, State) -> State;
+		 (proxy_protocol, State) -> State;
+		 ({sock_peer_name, _}, State) -> State;
 		 ({session_id, _}, State) -> State;
 		 (Opt, State) ->
 		      ?LOG_ERROR("Ignoring unknown option '~p'", [Opt]),
@@ -603,13 +700,14 @@ prepare_addr(T) when is_tuple(T) ->
 	    {error, einval}
     end.
 
+activate_socket(#state{sock = Sock, sock_mod = gen_tcp, shaper = none}) ->
+    inet:setopts(Sock, [{active, ?TCP_ACTIVE}]);
+activate_socket(#state{sock = Sock, sock_mod = SockMod, shaper = none}) ->
+    SockMod:setopts(Sock, [{active, ?TCP_ACTIVE}]);
+activate_socket(#state{sock = Sock, sock_mod = gen_tcp}) ->
+    inet:setopts(Sock, [{active, once}]);
 activate_socket(#state{sock = Sock, sock_mod = SockMod}) ->
-    case SockMod of
-	gen_tcp ->
-	    inet:setopts(Sock, [{active, once}]);
-	_ ->
-	    SockMod:setopts(Sock, [{active, once}])
-    end.
+    SockMod:setopts(Sock, [{active, once}]).
 
 cancel_timer(undefined) ->
     ok;
@@ -635,7 +733,7 @@ clean_treap(Treap, CleanPriority) ->
 	    Treap;
 	false ->
 	    {_Key, {TS, _}, _Value} = treap:get_root(Treap),
-	    if TS > CleanPriority ->
+	    if TS < CleanPriority ->
 		    clean_treap(treap:delete_root(Treap), CleanPriority);
 	       true ->
 		    Treap
@@ -645,13 +743,13 @@ clean_treap(Treap, CleanPriority) ->
 make_nonce(Addr, Nonces) ->
     Priority = now_priority(),
     {TS, _} = Priority,
-    Nonce = list_to_binary(integer_to_list(rand_uniform(1 bsl 32))),
-    NewNonces = clean_treap(Nonces, TS + ?NONCE_LIFETIME),
+    Nonce = integer_to_binary(rand_uniform(1 bsl 32)),
+    NewNonces = clean_treap(Nonces, TS - ?NONCE_LIFETIME),
     {Nonce, treap:insert(Nonce, Priority, Addr, NewNonces)}.
 
 have_nonce(Nonce, Nonces) ->
     TS = p1_time_compat:monotonic_time(micro_seconds),
-    NewNonces = clean_treap(Nonces, TS + ?NONCE_LIFETIME),
+    NewNonces = clean_treap(Nonces, TS - ?NONCE_LIFETIME),
     case treap:lookup(Nonce, NewNonces) of
 	{ok, _, _} ->
 	    {true, NewNonces};
@@ -672,6 +770,11 @@ check_integrity(User, Realm, Msg, [Pass | T]) ->
 	    check_integrity(User, Realm, Msg, T)
     end.
 
+check_expired_tag({expired, Pass}) ->
+    {Pass, true};
+check_expired_tag(Pass) ->
+    {Pass, false}.
+
 unmap_v4_addr({{0, 0, 0, 0, 0, 16#FFFF, D7, D8}, Port}) ->
     {{D7 bsr 8, D7 band 255, D8 bsr 8, D8 band 255}, Port};
 unmap_v4_addr(AddrPort) ->
@@ -707,18 +810,16 @@ get_sockmod(Opts, Sock) ->
 		true ->
 		    {ok, fast_tls};
 		false ->
-		    {ok, gen_tcp};
-		{error, _Reason} = Err ->
-		    Err
+		    {ok, gen_tcp}
 	    end
     end.
 
-get_certfile(Opts) ->
-    case catch iolist_to_binary(proplists:get_value(certfile, Opts)) of
-	Filename when is_binary(Filename), Filename /= <<"">> ->
-	    Filename;
-	_ ->
-	    undefined
+get_peername(Sock, Opts) ->
+    case proplists:get_value(sock_peer_name, Opts) of
+	{_, Addr} ->
+	    {ok, Addr};
+	undefined ->
+	    inet:peername(Sock)
     end.
 
 -ifdef(USE_OLD_INET_BACKEND).
@@ -735,18 +836,34 @@ is_tls_handshake({_, _, {_, Socket}}) ->
 	{ok, {_, _}} ->
 	    ?LOG_DEBUG("Determined transport protocol: TCP"),
 	    false;
-	{error, Reason} = Err ->
+	{error, Reason} ->
 	    ?LOG_INFO("Cannot determine transport protocol: ~s", [Reason]),
-	    Err
+	    false
     end.
 -endif.
 
-maybe_starttls(_Sock, fast_tls, undefined) ->
-    ?LOG_ERROR("Cannot start TLS connection: option 'certfile' is not set"),
-    {error, eprotonosupport};
-maybe_starttls(Sock, fast_tls, CertFile) ->
-    fast_tls:tcp_to_tls(Sock, [verify_none, {certfile, CertFile}]);
-maybe_starttls(Sock, gen_tcp, _CertFile) ->
+maybe_starttls(Sock, fast_tls, Opts) ->
+    case proplists:is_defined(certfile, Opts) of
+	true ->
+	    TLSOpts = lists:filter(
+			fun({certfile, _Val}) ->
+				true;
+			   ({dhfile, _Val}) ->
+				true;
+			   ({ciphers, _Val}) ->
+				true;
+			   ({protocol_options, _Val}) ->
+				true;
+			   (_Opt) ->
+				false
+			end, Opts),
+	    fast_tls:tcp_to_tls(Sock, [verify_none | TLSOpts]);
+	false ->
+	    ?LOG_ERROR("Cannot accept TLS connection: "
+		       "option 'certfile' is not set"),
+	    {error, eprotonosupport}
+    end;
+maybe_starttls(Sock, gen_tcp, _Opts) ->
     {ok, Sock}.
 
 prepare_response(State, Msg) ->
@@ -761,14 +878,16 @@ run_hook(HookName,
 		sock_mod = SockMod,
 		hook_fun = HookFun},
 	 #stun{'USERNAME' = User,
-	       'REALM' = Realm} = Msg)
+	       'REALM' = Realm,
+	       'ERROR-CODE' = Reason} = Msg)
   when is_function(HookFun) ->
     Info = #{id => ID,
 	     user => User,
 	     realm => Realm,
 	     client => Client,
 	     transport => stun_logger:encode_transport(SockMod),
-	     version => stun_codec:version(Msg)},
+	     version => stun_codec:version(Msg),
+	     reason => Reason},
     ?LOG_DEBUG("Running '~s' hook", [HookName]),
     try HookFun(HookName, Info)
     catch _:Err -> ?LOG_ERROR("Hook '~s' failed: ~p", [HookName, Err])
